import configparser


DEFAULT_INI = """[haunts]
# The Google Sheet Document id where you register events
# Required
CONTROLLER_SHEET_DOCUMENT_ID=<Google Sheet Document Id here>
# The sheet where to assign events category names to Google Calendar ids
# Default is "config"
# CONTROLLER_SHEET_NAME=config
# Events in the day start time in HH:MM format
# Default is 08:30
# START_TIME=08:30
"""

parser = configparser.RawConfigParser(allow_no_value=True)


def create_default(config):
    print("Creating default configuration")
    with open(config.resolve(), "w") as f:
        f.write(DEFAULT_INI)
    print("Created")


def init(config_file):
    global configuration
    with open(config_file.resolve(), "r") as config:
        parser.read_file(config)


def get(name, default=None):
    value = parser["haunts"].get(name, default)
    if value is None:
        raise KeyError(f"Not found: {name}")
    return default if value is None else value



import logging
import os

__author__ = """Luca Fabbri"""
__email__ = "l.fabbri@bopen.eu"
__version__ = "0.1.0"


LOGGER = logging.getLogger("haunts")

if os.environ.get("DEBUG"):
    LOGGER.setLevel(logging.DEBUG)



import sys
import time
import string
import datetime
import click

from googleapiclient.errors import HttpError
from googleapiclient.discovery import build
from google_auth_oauthlib.flow import InstalledAppFlow
from google.auth.transport.requests import Request
from google.oauth2.credentials import Credentials

from .ini import get
from . import actions
from .calendars import create_event, delete_event, ORIGIN_TIME

# If modifying these scopes, delete the sheets-token file
SCOPES = ["https://www.googleapis.com/auth/spreadsheets"]
creds = None


def get_col(row, index):
    try:
        return row[index]
    except IndexError:
        return None


def get_credentials(config_dir):
    global creds
    if creds is not None:
        return
    # The token stores the user's access and refresh tokens, and is
    # created automatically when the authorization flow completes for the first
    # time.
    token = config_dir / "sheets-token.json"
    credentials = config_dir / "credentials.json"
    if token.is_file():
        creds = Credentials.from_authorized_user_file(token.resolve(), SCOPES)
    # If there are no (valid) credentials available, let the user log in.
    if not creds or not creds.valid:
        if creds and creds.expired and creds.refresh_token:
            creds.refresh(Request())
        else:
            flow = InstalledAppFlow.from_client_secrets_file(
                credentials.resolve(), SCOPES
            )
            creds = flow.run_local_server(port=0)
        # Save the credentials for the next run
        with open(token.resolve(), "w") as token_file:
            token_file.write(creds.to_json())


def get_headers(sheet, month, indexes=False):
    """Scan headers of a month and returns a structure that assign headers names to indexes"""
    selected_month = (
        sheet.values()
        .get(spreadsheetId=get("CONTROLLER_SHEET_DOCUMENT_ID"), range=f"{month}!A1:ZZ1")
        .execute()
    )
    values = selected_month["values"][0]
    if indexes:
        return {k: values.index(k) for k in values}
    return {k: string.ascii_lowercase.upper()[values.index(k)] for k in values}


def sync_events(config_dir, sheet, data, calendars, days, month):
    """Enumerate every data in the sheet.
    Create an event when action column is empty
    """
    headers = get_headers(sheet, month)
    headers_id = get_headers(sheet, month, indexes=True)
    last_to_time = None
    last_date = None

    for y, row in enumerate(data["values"]):
        current_date = get_col(row, headers_id["Date"])
        date = ORIGIN_TIME + datetime.timedelta(days=current_date)

        default_start_time = (
            get_col(row, headers_id["Start time"])
            if headers_id.get("Start time") and get_col(row, headers_id["Start time"])
            else None
        )

        # In case we changed day, let's restart from START_TIME
        if current_date != last_date:
            last_to_time = None
        last_date = current_date

        # short circuit for date filters
        skip = len(days) > 0
        for d in days:
            if date.date() == d.date():
                skip = False
                break
        if skip:
            continue

        calendar = None
        try:
            calendar = calendars[get_col(row, headers_id["Project"])]
        except KeyError:
            click.echo(
                f"Cannot find a calendar id associated to project \"{get_col(row, headers_id['Project'])}\""
            )
            sys.exit(1)

        try:
            action = row[headers_id["Action"]]
            if action == actions.IGNORE:
                continue
            if action == actions.DELETE:
                delete_event(
                    config_dir=config_dir,
                    calendar=calendar,
                    event_id=get_col(row, headers_id["Event id"]),
                )
                click.echo(f'Deleted event "{get_col(row, headers_id["Activity"])}"')
                request = sheet.values().batchClear(
                    spreadsheetId=get("CONTROLLER_SHEET_DOCUMENT_ID"),
                    body={
                        "ranges": [
                            f"{month}!{headers['Event id']}{y + 2}",
                            f"{month}!{headers['Link']}{y + 2}",
                            f"{month}!{headers['Action']}{y + 2}",
                        ],
                    },
                )

                try:
                    request.execute()
                except HttpError as err:
                    if err.status_code == 429:
                        click.echo("Too many requests")
                        click.echo(err.error_details)
                        click.echo("haunts will now pause for a while ⏲…")
                        time.sleep(60)
                        click.echo("Retrying…")
                        request.execute()
                    else:
                        raise

                continue
            if action:
                # There's something in the action cell, but not recognized
                click.echo(f"Unknown action {action}. Ignoring…")
                continue
        except IndexError:
            # We have no data there
            pass

        event = create_event(
            config_dir=config_dir,
            calendar=calendar,
            date=date,
            summary=get_col(row, headers_id["Activity"]),
            details=get_col(row, headers_id["Details"]),
            length=get_col(row, headers_id["Spent"]),
            from_time=default_start_time or last_to_time,
        )
        last_to_time = event["next_slot"]

        request = sheet.values().batchUpdate(
            spreadsheetId=get("CONTROLLER_SHEET_DOCUMENT_ID"),
            body={
                "valueInputOption": "USER_ENTERED",
                "data": [
                    # Put the action to actions.IGNORE, in this way it will not be processed again
                    {
                        "range": f"{month}!{headers['Action']}{y + 2}",
                        "values": [[actions.IGNORE]],
                    },
                    # Save the event id, required to interact with the event in future
                    {
                        "range": f"{month}!{headers['Event id']}{y + 2}",
                        "values": [[event["id"]]],
                    },
                    # Quick link to the event on the calendar
                    {
                        "range": f"{month}!{headers['Link']}{y + 2}",
                        "values": [[f"=HYPERLINK(\"{event['link']}\";\"open\")"]],
                    },
                ],
            },
        )

        try:
            request.execute()
        except HttpError as err:
            if err.status_code == 429:
                click.echo("Too many requests")
                click.echo(err.error_details)
                click.echo("haunts will now pause for a while ⏲…")
                time.sleep(60)
                click.echo("Retrying…")
                request.execute()
            else:
                raise
    click.echo("Done!")


def get_calendars(sheet):
    RANGE = f"{get('CONTROLLER_SHEET_NAME', 'config')}!A2:B"
    calendars = (
        sheet.values()
        .get(spreadsheetId=get("CONTROLLER_SHEET_DOCUMENT_ID"), range=RANGE)
        .execute()
    )
    values = calendars.get("values", [])
    return {alias: id for [id, alias] in values}


def sync_report(config_dir, month, days=[]):
    """Open a sheet, analyze it and populate calendars with new events"""
    # The ID and range of the controller timesheet
    get_credentials(config_dir)
    service = build("sheets", "v4", credentials=creds)

    # Call the Sheets API
    sheet = service.spreadsheets()

    click.echo("Started calendars synchronization")

    try:
        document_id = get("CONTROLLER_SHEET_DOCUMENT_ID")
    except KeyError:
        click.echo(
            "A value for CONTROLLER_SHEET_DOCUMENT_ID is required but "
            "is not specified in your ini file"
        )
        sys.exit(1)

    data = (
        sheet.values()
        .get(
            spreadsheetId=document_id,
            range=f"{month}!A2:ZZ",
            valueRenderOption="UNFORMATTED_VALUE",
        )
        .execute()
    )

    calendars = get_calendars(sheet)
    sync_events(config_dir, sheet, data, calendars, days=days, month=month)



import sys
import time
import datetime
import click
from dateutil import parser

from googleapiclient.errors import HttpError
from googleapiclient.discovery import build
from google_auth_oauthlib.flow import InstalledAppFlow
from google.auth.transport.requests import Request
from google.oauth2.credentials import Credentials

from . import LOGGER
from .ini import get

LOCAL_TIMEZONE = datetime.datetime.utcnow().astimezone().strftime("%z")
# Weird google spreadsheet date management
ORIGIN_TIME = datetime.datetime.strptime(
    f"1899-12-30T00:00:00{LOCAL_TIMEZONE}", "%Y-%m-%dT%H:%M:%S%z"
)
# If modifying these scopes, delete the calendars-token file.
SCOPES = ["https://www.googleapis.com/auth/calendar"]
creds = None


def formatDate(date, format):
    return parser.isoparse(date).strftime(format)


def get_credentials(config_dir):
    global creds
    if creds is not None:
        return
    # The file token.json stores the user's access and refresh tokens, and is
    # created automatically when the authorization flow completes for the first
    # time.
    token = config_dir / "calendars-token.json"
    credentials = config_dir / "credentials.json"
    if not credentials.exists():
        click.echo(
            f"Missing credentials file at {credentials.resolve()}. "
            f"Did you created a Google Cloud project and downloaded the credentials file?"
        )
        sys.exit(1)
    if token.is_file():
        creds = Credentials.from_authorized_user_file(token.resolve(), SCOPES)
    # If there are no (valid) credentials available, let the user log in.
    if not creds or not creds.valid:
        if creds and creds.expired and creds.refresh_token:
            creds.refresh(Request())
        else:
            flow = InstalledAppFlow.from_client_secrets_file(
                credentials.resolve(), SCOPES
            )
            creds = flow.run_local_server(port=0)
        # Save the credentials for the next run
        with open(token.resolve(), "w") as token:
            token.write(creds.to_json())


def init(config_dir):
    get_credentials(config_dir)


def create_event(config_dir, calendar, date, summary, details, length, from_time=None):
    get_credentials(config_dir)
    service = build("calendar", "v3", credentials=creds)

    from_time = from_time or get("START_TIME", "08:30")
    start = datetime.datetime.strptime(
        f"{date.strftime('%Y-%m-%d')}T{from_time}:00{LOCAL_TIMEZONE}",
        f"%Y-%m-%dT%H:%M:%S%z",
    )

    startParams = None
    endParams = None
    haveLength = length is not None and type(length) is not str
    duration = None
    if haveLength:
        duration = float(length)
        delta = datetime.timedelta(hours=duration)
    else:
        delta = datetime.timedelta(hours=0)
    end = start + delta

    if haveLength:
        startParams = {
            "dateTime": start.isoformat(),
        }
        endParams = {
            "dateTime": end.isoformat(),
        }
    else:
        startParams = {
            "date": start.isoformat()[:10],
        }
        endParams = {
            "date": (end + datetime.timedelta(days=1)).isoformat()[:10],
        }

    event_body = {
        "summary": summary,
        "description": details,
        "start": startParams,
        "end": endParams,
    }

    def execute_creation():
        LOGGER.debug(calendar, date, summary, details, length, event_body, from_time)
        event = service.events().insert(calendarId=calendar, body=event_body).execute()
        return event

    try:
        event = execute_creation()
    except HttpError as err:
        if err.status_code == 429:
            click.echo("Too many requests")
            click.echo(err.error_details)
            click.echo("haunts will now pause for a while ⏲…")
            time.sleep(60)
            click.echo("Retrying…")
            event = execute_creation()
        else:
            raise

    LOGGER.debug(event.items())
    if duration:
        click.echo(
            f'Created event "{summary}" from {formatDate(event["start"]["dateTime"], "%H:%M")} '
            f'to {formatDate(event["end"]["dateTime"], "%H:%M")} ({duration}h) '
            f'in date {formatDate(event["start"]["dateTime"], "%d/%m")} '
            f'on calendar {event["organizer"]["displayName"]}'
        )
    else:
        click.echo(
            f'Created event "{summary}" (full day) '
            f'in date {formatDate(event["start"]["date"], "%d/%m")} '
            f'on calendar {event["organizer"]["displayName"]}'
        )

    event_data = {
        "id": event["id"],
        "next_slot": end.strftime("%H:%M") if haveLength else from_time,
        "link": event["htmlLink"],
    }
    return event_data


def delete_event(config_dir, calendar, event_id):
    get_credentials(config_dir)
    service = build("calendar", "v3", credentials=creds)
    if not event_id:
        click.echo(f"Missing id. Skipping…")
        return
    try:
        service.events().delete(calendarId=calendar, eventId=event_id).execute()
    except HttpError as err:
        if err.status_code == 410:
            click.echo(f"Event {event_id} already deleted")



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

requirements = [
    "Click>=7.0",
    "python-dateutil",
    "google-api-python-client",
    "google-auth-httplib2",
    "google-auth-oauthlib",
    "google-auth<2dev",
]

test_requirements = []

setup(
    author="Luca Fabbri",
    author_email="l.fabbri@bopen.eu",
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    description="Fill and sync Google Calendars with events taken from a Google spreadsheet",
    entry_points={
        "console_scripts": [
            "haunts=haunts.cli:main",
        ],
    },
    install_requires=requirements,
    license="GNU General Public License v3",
    long_description=readme + "\n\n" + history,
    include_package_data=True,
    keywords="google-calendar spreadsheet reports",
    name="haunts",
    packages=find_packages(include=["haunts", "haunts.*"]),
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/bopen/haunts",
    version="0.3.0",
    zip_safe=False,
)



"""Console script for haunts."""
import datetime
import os
import sys
from pathlib import Path

import click

from .ini import create_default, init

from .calendars import init as init_calendars
from .spreadsheet import sync_report


@click.command()
@click.argument("sheet", required=False)
@click.option(
    "--day",
    "-d",
    multiple=True,
    help='day filter in format "YYYY-MM-DD". Can be provided multiple times.',
)
@click.option(
    "--config",
    "-c",
    "run_configuration",
    help="configure haunts working folder and .ini file.",
    is_flag=True,
    show_default=True,
    default=False,
)
def main(sheet=None, day=[], run_configuration=False):
    """
    Sync events from a Google Sheet to your Google Calendar.
    """

    # config phase
    config_dir = Path(os.path.expanduser("~/.haunts"))

    if not run_configuration and not config_dir.is_dir():
        click.echo(
            f"Configuation directory at {config_dir.resolve()} not found. Use --config  to create it."
        )
        sys.exit(1)

    config = Path(os.path.expanduser(f"{config_dir.resolve()}/haunts.ini"))
    if not run_configuration and not config_dir.is_dir():
        click.echo(
            f"Configuation file at {config.resolve()} not found. Use --config to create it."
        )
        sys.exit(1)
    if run_configuration and not config_dir.is_dir():
        click.echo(f"Creating config directory at {config_dir.resolve()}")
        config_dir.mkdir()
        click.echo("…created")
        if not config.is_file():
            create_default(config)
            click.echo(
                f"Manage you settings at {config.resolve()} before running haunts again."
            )
            sys.exit(1)

    if not run_configuration and not sheet:
        click.echo(f"Argument SHEET is required if no '--config' flag is provided.")
        sys.exit(1)

    init(config)

    if run_configuration:
        click.echo("All done. You can now start using haunts.")
        sys.exit(0)

    init_calendars(config_dir)
    sync_report(
        config_dir,
        sheet,
        days=[datetime.datetime.strptime(d, "%Y-%m-%d") for d in day],
    )
    return 0


if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover
